//                              -*- Mode: C++ -*-
// optimizing.cc --- Addition to the read API to do binary search
// Author          : Tim Lethbridge
// Created On      : Thu Feb 18 17:40:10 1998
//

// This file contains optimized versions of code that was originally
// in read-API.cc. It is placed here to avoid changing the shared object.

# include "optimizing.hh"
# include <string.h>

template <class TOut>
class BSVectorIterator: public BaseIterator<Vector<TOut> *,TOut> {
// Iterator on a vector from container.hh that looks for the first occurrence
// of an initial substring in strings in the vector and then returns every
// matching element
// **************************************************
protected:
  // the pattern to search for
  char *_initpat;

  // a function which takes a TOut and returns a char *
  // used for the binary search
  char *(*_getStr)(TOut);

  TOut _nextComputed;
  int _nextVectorIndex; // points 1 past _nextComputed
  int _vectorSize;
  int _initpatSize;
  bool _doneBinarySearch; // false initially


  inline void computeOneAhead()
  // --------------------------------------------------
  {
    if( ! _doneBinarySearch) {
       // do binary search for location start
       // if not found, sets _eoi to true
       // if found sets _nextComputed and _nextVectorIndex
       computeBinary();
       _doneBinarySearch = true;
    } /* initial binary search */

    else {
      if(_nextVectorIndex >= _vectorSize) {
         _eoi = true;
      }
      else {
         _nextComputed =(*_iterationBase)[_nextVectorIndex++];
         if(strncmp(_getStr(_nextComputed), _initpat, _initpatSize)) {
           // We are past the region where the pattern occurs
           _eoi = true;
         } /* past region */
      } /* not past end of vector */
    } /* sequential following binary search */
  } /* computeOneAhead() */


  inline TOut computeNext()
  // --------------------------------------------------
  {
      TOut tmp;
      tmp = _nextComputed;
      computeOneAhead(); // sets _eoi
      return(tmp);
  }

  inline void computeBinary()
  // --------------------------------------------------
  // This can be overridden to provide special purpose binary search
  {
      int prev;
      int first = -1;
      int last = _vectorSize +1;
      int comparison;

      _nextVectorIndex = (last + first)/2;
      _nextComputed =(*_iterationBase)[_nextVectorIndex];
      comparison = strncmp(_getStr(_nextComputed), _initpat,
         _initpatSize);
      while(first < last)
      {
         if(comparison >= 0) last = _nextVectorIndex;
         else first = _nextVectorIndex;
         prev = _nextVectorIndex;
         _nextVectorIndex = (last + first)/2;
         if(prev == _nextVectorIndex) {
           /* terminating condition */
           _nextVectorIndex++;
           _nextComputed =(*_iterationBase)[_nextVectorIndex];
           comparison = strncmp(_getStr(_nextComputed), _initpat,
              _initpatSize);
           last = first;
         }
         _nextComputed = (*_iterationBase)[_nextVectorIndex];
         comparison = strncmp(_getStr(_nextComputed), _initpat,
            _initpatSize);
      }
      if(comparison != 0) {
         // Failed to find the substring (or not sorted)
         _eoi = true;
      }
      else {
         // Found start of vector items that start with the substring
         _nextVectorIndex++;
      }
  }

public:
  BSVectorIterator(Vector<TOut> *base, char *initpat, char
      *(getStr)(TOut)) :
  // --------------------------------------------------
      BaseIterator<Vector<TOut>*,TOut>(base) {
      _initpat = (char *)malloc(strlen(initpat)+1);
      strcpy(_initpat, initpat);
      _getStr = getStr;
      recursiveDelete(false);
      _vectorSize = _iterationBase->size();
      _nextVectorIndex = 0;
      _doneBinarySearch = false;
      _initpatSize = strlen(_initpat);

      if(_vectorSize == 0) {
        _eoi=true;
      }
      else {
        _eoi = false;
        computeOneAhead();
      }
  };

  ~BSVectorIterator() { free(_initpat); };

}; /* class BinarySearchVectorIterator */




int findConstStartOfRegEx(char *regexp, char *constPart)
{
   /* find the constant initial portion of a regular expression */
   /* side-effects constPart by writing the constant portion to it */
   /* returns the length of constPart */
   char *currentChar, *currentCharOut;
   int count;

   *constPart = '\0';
   count = 0;

   /* Null case -- no initial string */
   if(*regexp != '^') return count;

   /* loop through constant characters */
   currentChar = regexp+1;
   currentCharOut = constPart;
   while(*currentChar != '\0')
   {
     /* special characters always terminate constant portion */
     if(*currentChar == '.' || *currentChar == '[' || *currentChar == '*')
        return count;

     /* backslash followed by a non-special character has a special
        meaning */
     if(*currentChar == '\\')
     {
        currentChar++;
        if(*currentChar != '.' && *currentChar != '*' &&
          *currentChar != '[' && *currentChar != '\\' &&
          *currentChar != '$')
          return count;
     }

     /* Dollar sign at the end has special meaning */
     if(*currentChar == '$' && *(currentChar+1) == '\0')
       return count;        

     /* at this point we know we have a valid constant character */
     *currentCharOut = *currentChar;
     currentCharOut++;
     currentChar++;
     count++;
     *currentCharOut = '\0';
  }
  return count;
}

Iterator *cdbGetObjectListOptRegExp(char *regexp, class_t filterClass)
//--------------------------------------------------
{
  VoidIterator<SoftwareObject*> *SOiter;
  char startstring[400];
  int startStringlen=0;

  // If the regular expression starts with a ^ and a non-special
  // character, then we can do an efficient binary search to find an
  // appropriate starting point

  // Find the starting sequence of characters that the regexp
  // will match.

  startStringlen = findConstStartOfRegEx(regexp, startstring);

  /* Use the regular function if we need to */
  if(startStringlen == 0)
     return(cdbGetObjectListRegExp(regexp, filterClass));

  /* otherwise use an iterator that only works on a subsequence of
     the objects defined by the BSVectorIterator */
  SOiter = new BSVectorIterator<SoftwareObject*>(the_DB->allObjects(),
     startstring, cdbGetObjectName);
 
  return( new RegexpIterator<SoftwareObject*>(
                new TypeFilterIterator<SoftwareObject*,SoftwareObject*>(
                  SOiter, filterClass),
                        regexp, cdbGetObjectName));
}  /* cdbGetObjectListOptRegExp() */



